#!/usr/bin/python
#=========================================================================
# SCOPE:
SCOPE = 'ftpDataFiles.py looks for new or updated data files and '+\
        '  transmits them to a remote server using FTP. '

#
# 
# Written by:  Dan Robinson, Remote Sensing Solutions
# Date: January 2010
#
# Version Number
versionNumber = '2.0'
#
# Modifications
# (list version number, author, date and modifications)
#
# Version 2.0 
# June 2010
# Dan Robinson
# - Made Mission ID optional
# - Modified format of configuration file
# - Override default file paths via config file
#
#=========================================================================
# Import libraries
# - Standard modules that should come with standard Python install
import ftplib, sys, os, string, time, gzip, shutil, traceback

#---------------------------------------------------------------------------
# Build file list function
# - Callback function.  Called by FTP object upon returning list of files.
#   This function is called iteratively, as file names are returned one 
#   at a time.
#
# Parameters:
#   - fileName:  data returned from call to retrlines
#
def buildFileList(fileName):
    serverFiles.append(fileName)

#---------------------------------------------------------------------------
# Write status File function
# - Writes status to a log file
#
# Parameters:
#   - statusFilePath: name of the status file that is being created
#   - statusString:  string describing status that has occurred
#
def writeStatusFile( statusFilePath, statusString ):
    if not (os.path.exists( statusFilePath ) ):
        statusLog = open( statusFilePath, 'w' )
    else:
    	statusLog = open( statusFilePath, 'a' )
        
    statusLog.write( statusString )
    statusLog.close( )
    
#---------------------------------------------------------------------------
# Compresses a file using gzip
#
# Parameters:
#   - uncompressedFilePath: path to the file that is to be zipped
#
def compressFile( uncompressedFilePath ):
    
    # create the output zipped file
    compressedFile = open( uncompressedFilePath+'.gz', "wb" )
    
    # compression level 6 seems to be the best tradeoff of run time vs file size
    dest= gzip.GzipFile( compresslevel=6, fileobj=compressedFile, mode='wt' )
    source = open( uncompressedFilePath, 'rb' )
    
    #data blocks of 4k seem to be most efficient (faster than 1k, 2k, 8k)
    block= source.read( 4096 )
    while block:
        dest.write(block)
        block= source.read( 4096 )

    # clean up open objects
    dest.flush()
    dest.close()
    compressedFile.close()
    source.close()
    
# format exceptions in a more readable output
def formatExceptionInfo(maxTBlevel=5):
     cla, exc, trbk = sys.exc_info()
     excName = cla.__name__
     try:
         excArgs = exc.__dict__["args"]
     except KeyError:
         excArgs = "<no args>"
     excTb = traceback.format_tb(trbk, maxTBlevel)
     return (excName, excArgs, excTb)

#---------------------------------------------------------------------------
# Setup varibles and file system
#-------------------------------
# Local Variables
sleepTime = 1.         # time to wait before checking file size again
newFileWait = 60       # time to wait before checking directory for new files

print sys.argv[0]

# Command Line Argument Syntax
syntax = '\n'
syntax += 'SYNTAX: \n'
syntax += '\n'
syntax += '-> ftpDataFiles.py <-m [mission ID]> \n '
syntax += '                   <-h or --help>\n'
syntax += '                   <-v or --version>\n'
syntax += '                   <-i [iterations]>\n'
syntax += '\n\n\n OPTIONS:\n\n'
syntax += '  -m [mission ID]:    Mission ID. This is used to define file directories\n'
syntax += '                      when using the default file paths.\n'
syntax += '  -h or --help:       prints to stdout syntax and exits\n'
syntax += '  -v or --version:    print version number exits\n'
syntax == '  -noprefix:          Do not put a prepend the current time on file names.\n'
syntax += '  -i [iterations]:    Sets the number of iterations to check for file growth\n'
syntax += '                      before determining that file has stopped growing\n'
syntax += '                      where one iteration = one second. Default = 30.\n'

print 'Running...'

exitNow = 0
iterLimit = 10 
missionIDFound = 0
missionID = ''
skip = 0
prefix = True

# Check for arguments
for i in range(len(sys.argv)-1):
    if skip:
        skip = 0
        continue
    
    if sys.argv[i+1] == '-h' or sys.argv[i+1] == '--help':
        print syntax
        exitNow = 1
        
    elif sys.argv[i+1] == '-v' or sys.argv[i+1] == '--version':
        print '\n ftpDataFiles.py written by Remote Sensing Solutions'
        print '      Version: '+version+' \n'
        exitNow = 1
        
    elif sys.argv[i+1] == '-verbose':
        verboseMode = 1
        print 'verbose option set.  Program will print diagnostic information.\n'
            
    elif sys.argv[i+1] == '-noprefix':
        prefix = False
        print 'Not using prefix on file names.\n'
            
    elif sys.argv[i+1] == '-i':
        iterLimit = int(sys.argv[i+2])
        skip = 1
    
    elif sys.argv[i+1] == '-m':
        missionID = sys.argv[i+2]
        skip = 1
    
    else:
        print '\n Incorrect Syntax! \n'+syntax
        exitNow = 1

if exitNow == 1: 
    time.sleep( 15 )
    sys.exit()
    
    
# Assign default directories
parentDirectory = os.path.expanduser("~") + '/data.links'
serverDirectory = '/hrd/incoming/gamache/test/' + missionID
dataDirectory = os.path.join(parentDirectory, serverDirectory)

# Get FTP info
# Open Underlying File
zippedFile = open('ftpInfo', mode='rb')

# Create the uncompressing wrapper for the zipped file
# This allow us to treat the file as if it was not zipped
ftpInfoFile = gzip.GzipFile(fileobj=zippedFile, mode='rt')
ftpInfoLines = ftpInfoFile.readlines()
ftpInfoFile.close()
zippedFile.close()

wsFound = False
portFound = False
unFound = False
pwFound = False
pzFound = False
zip = True

for ftpInfo in ftpInfoLines:
    ftpInfo = ftpInfo.strip('\n')
    ftpInfo = ftpInfo.split()
    
    if (len(ftpInfo) < 2):
        continue
    
    if not ftpInfo[0].startswith('\\'):
        continue
    
    if ftpInfo[0] == '\\Address':
        webSite  = ftpInfo[1]
        wsFound = True
    elif ftpInfo[0] == '\\Port':
        port     = ftpInfo[1]
        portFound = True
    elif ftpInfo[0] == '\\Username':
        username = ftpInfo[1]
        unFound = True
    elif ftpInfo[0] == '\\Password':
        password = ftpInfo[1]
        pwFound = True
    elif ftpInfo[0] == '\\Prezipped':
        pzFound = True
        if ftpInfo[1] == "yes":
            zip = False
        else:
            zip = True
    elif ftpInfo[0] == '\\Path':
        dataDirectory = ftpInfo[1]

logDirectory  = os.path.join(dataDirectory, 'log')

if not wsFound:
    print 'No Address provided for FTP destination.'
if not portFound:
    print 'No port number provided for FTP destination.'
if not unFound:
    print 'No username provided for FTP destination.'
if not pwFound:
    print 'No password provided for FTP destination.'
if not pzFound:
    print 'Prezipped keyword not present in configuration.'
    
if not wsFound or not portFound or not unFound or not pwFound or not pzFound:
    print 'Correct configuration and try again. Exiting.'
    sys.exit()

looping = True
fileCount = 0
timeStamp = time.strftime('%Y%m%d_%H%M%S')
statusFileName = 'status_' + timeStamp + '.log'
statusFilePath = logDirectory + '/' + statusFileName

# If directory is not there, exit.
if not(os.path.exists(dataDirectory)):
    print dataDirectory + 'does not exist, creating.'
    os.makedirs( dataDirectory )

# Check for log file directory
if not ( os.path.exists(logDirectory) ):
    # Doesn't exist, create it
    os.makedirs( logDirectory )

writtenFiles = []
if (os.path.isfile('writtenFiles.list')):
    writtenFileList = open( 'writtenFiles.list', 'r' )
    writtenFiles = writtenFileList.readlines()
    
    # File names have carriage return appended
    i = 0
    while i < len( writtenFiles ):
        writtenFiles[i] = writtenFiles[i].rstrip()
        i += 1
    writtenFileList.close()

while looping:  
    
    # Retrieve current file list and sort it to find the most recent file
    currentDirList = os.listdir( dataDirectory )
    currentFileList = []
    for thisEntry in currentDirList:
        if ( os.path.isfile(dataDirectory+'/'+thisEntry) ):
            # If zipping files before transmitting
            # only include files not already zipped
            if zip and (not (thisEntry.endswith('.gz'))):
                currentFileList.append( thisEntry )
            # if not zipping, only include files that are already zipped
            elif not zip and thisEntry.endswith('.gz'):
                currentFileList.append( thisEntry )
    currentFileList.sort()
    if ( len(currentFileList) == 0 ):
        print 'No files to process. Waiting %d seconds to check again.' % newFileWait
        time.sleep( newFileWait )
        continue
    monitoredFile = currentFileList[len(currentFileList)-1]
    if monitoredFile not in writtenFiles:
        monitoredFileSize = os.stat(dataDirectory+'/'+monitoredFile).st_size
        iterations = 1
        print 'file', monitoredFile, 'has size ', monitoredFileSize
        time.sleep( sleepTime )
        # Loop, checking file size and number of iterations since last change
        while( iterations < iterLimit ):
            fileSize = os.stat(dataDirectory+'/'+monitoredFile).st_size
            if ( fileSize > monitoredFileSize ):
                iterations = 0
                monitoredFileSize = fileSize
                print 'file', monitoredFile, 'has size ', monitoredFileSize
                sleepTime = 1
            else:
                print 'file did not grow %d' % iterations
                iterations += 1
                #sleepTime = 60
                sleepTime = 1
            time.sleep( sleepTime )
        
        if zip:
            # Zip the file for more efficient transmission
            print 'Compressing ' + monitoredFile
            compressFile( dataDirectory+'/'+monitoredFile )
            print 'File size for %s = %d' % (monitoredFile, fileSize)
            transmitFile += '.gz'
            print 'File size for %s = %d' % (monitoredFile+'.gz', \
                                             os.stat(transmitFile).st_size)
                
        # Put prefix on file if appropriate
        if prefix:
            filePrefix = 'CT_%s_%s_' % (time.strftime('%Y%m%d'), time.strftime('%H%M%S'))
            transmitFile = filePrefix + monitoredFile
            shutil.copyfile(os.path.join(dataDirectory, monitoredFile), os.path.join(dataDirectory, transmitFile))
        else:
            transmitFile = monitoredFile 
        print transmitFile
            
        # Create and open FTP session
        try:
            f = ftplib.FTP( )
            f.connect( webSite, port )
        except ftplib.all_errors, error:
            statusString = 'Cannot connect. Error: %s\n' % error
            print 'ERROR:', statusString
            writeStatusFile( statusFilePath, statusString )
            #if prefix:
            #    os.remove( os.path.join(dataDirectory,transmitFile) )
                        
        else:
                
            try:
                f.login( username, password )
            except ftplib.all_errors, error:
                statusString = 'Login failed. Error: %s\n' % error
                print 'ERROR:', statusString
                writeStatusFile( statusFilePath, statusString )
                #if prefix:
                #    os.remove( os.path.join(dataDirectory,transmitFile) )
                        
            else:
                # Change to directory to which files will be transferred
                try:
                    f.mkd( serverDirectory )
                except ftplib.all_errors, error:
                    # This is actually good, do nothing
                    # There has to be a better way to do this...
                    x = 0
                    
                try:
                    f.cwd( serverDirectory )
                except ftplib.all_errors, error:
                    statusString = 'Directory ' + dataDirectory + ' Error: %s\n' % error
                    print 'ERROR:', statusString
                    writeStatusFile( statusFilePath, statusString )
                else:
                    command = 'STOR ' + os.path.join(serverDirectory, transmitFile)
                    try:
                        print 'Sending file: ' + os.path.join(serverDirectory, transmitFile)
                        f.storbinary( command, open( os.path.join(dataDirectory,transmitFile), 'rb' ) )
                        statusString = 'Sent: ' + transmitFile
                        print statusString
                        writeStatusFile( statusFilePath, statusString+'\n' )
                        writtenFileList = open( 'writtenFiles.list', 'a+')
                        writtenFileList.write( monitoredFile+'\n' )
                        writtenFileList.close( )
                        fileCount += 1
                        # If a prefix was added, delete the copy that was made
                        #if prefix:
                        #    os.remove( os.path.join(dataDirectory,transmitFile) )
                    except ftplib.all_errors, error:
                        # If a prefix was added, delete the copy that was made
                        #if prefix:
                        #    os.remove( os.path.join(dataDirectory,transmitFile) )
                        statusString = 'File ' + transmitFile + ' Error: %s' % error
                        statusString += '\n'
                        writeStatusFile( statusFilePath, statusString )
                        print formatExceptionInfo()
                    
            # Write status to log file
            statusString = '%d files written on %s\n' % ( fileCount, time.asctime() )
            writeStatusFile( statusFilePath, statusString )
            
            writtenFiles.append( monitoredFile )
                                    
            # Terminate the FTP session
            f.quit( )
            
        # Delete prefix file
        if prefix:
                os.remove( os.path.join(dataDirectory,transmitFile) )

    else:
        # monitored file was already written, wait then look for new files
        print 'No new files to process at this time. Waiting %d to check again.' % newFileWait
        time.sleep( newFileWait )
print 'Execution complete.  Script will exit in', sleepTime, 'seconds.'
time.sleep( sleepTime )

print 'Done.'        
